//函数 mystrncmp(const char* str1, const char* str2, unsigned int n) 的验证
//要点：最多只比较字符串 str1 和 str2 前 n 个字符的大小。
//关键字：字符串、内置谓词、逻辑变量

#include <limits.h>

///////////////////////////////////////////逻辑变量////////////////////////////////////////////////
//@ logic char *p1, *p2;      // p1和p2分别记录函数入口处字符串str1和str2的地址
//@ logic unsigned int oldn;  // 记录形参n的值
/////////////////////////////////////////mystrncmp2函数的验证/////////////////////////////////////////
// 函数协议：
//前条件：函数入口处str1和str2分别指向两个物理字符串；给出n的取值范围；给出字符的取值范围。
//后条件：函数返回值=0，则str1所指字符串和str2所指字符串前n个字符相同；
//           返回值>0，则str1所指字符串的前n个字符大于str2所指字符串的前n个字符；
//           返回值<0，则str1所指字符串的前n个字符小于str2所指字符串的前n个字符。 
//       若str1和str2的串长都大于n，则只比较前n个字符（即比较str1和str2长为n的前缀）。

/*@ 
  requires 
      \is_pstring(str1, \strlen(p1)) && \is_pstring(str2, \strlen(p2)) 
      && p1 == str1 && p2 == str2 
      && oldn == n && n >= 0 && n < INT_MAX 
      && (\forall integer i:[0..n-1]. p1[i]>=0 && p1[i]<=128 && p2[i]>=0 && p2[i]<=128)
      ;
  
  ensures 
      \is_pstring(p1, \strlen(p1)) && \is_string(p2, \strlen(p2)) 
      && 
       (
        (\strlen(p1) >= oldn && \strlen(p2) >= oldn) &&  // 2个串长均大于n
        (
           \prefix(\string(p1, \strlen(p1)), oldn) < \prefix(\string(p2, \strlen(p2)), oldn) && \result < 0
          ||
           \prefix(\string(p1, \strlen(p1)), oldn) == \prefix(\string(p2, \strlen(p2)), oldn) && \result == 0
          ||
           \prefix(\string(p1, \strlen(p1)), oldn) > \prefix(\string(p2, \strlen(p2)), oldn) && \result > 0
        )    
      ||
        (\strlen(p1) < oldn || \strlen(p2) < oldn) &&  // 有串长小于n
        (
          \string(p1, \strlen(p1)) < \string(p2, \strlen(p2))  && \result < 0
          ||
          \string(p1, \strlen(p1)) == \string(p2, \strlen(p2)) && \result == 0
          ||
          \string(p1, \strlen(p1)) > \string(p2, \strlen(p2))  && \result > 0
        )
       )
      ;
*/
int mystrncmp(const char* str1,const char* str2, unsigned int n) {
    int j = 0;
    /*@ 
      loop invariant
          oldn == n && n >= 0  && j >= 0 && j <= n && n < INT_MAX 
          && j<=\strlen(p1)&& j<=\strlen(p2) 
          && str1 == p1 + j && str2 == p2 + j && \is_pstring(p1, \strlen(p1)) 
          && \is_pstring(p2, \strlen(p2)) 
          && (\forall integer i:[0..j-1]. p1[i] == p2[i]) 
          && (\forall integer i:[0..n-1]. p1[i]>=0 && p1[i]<=128 && p2[i]>=0 && p2[i]<=128)
          ;
    */
    while(j < n && *str1 == *str2){
        if(*str1 == '\0'){
            return 0;
        }
        str1++; str2++; j++;
    }
    if(n == 0 || j == n){
        return 0;
    } else {
        return *str1 - *str2;
    }
}